#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['py_linq']

package_data = \
{'': ['*'],
 'py_linq': ['.pytest_cache/*',
             '.pytest_cache\\v/*',
             '.pytest_cache\\v\\cache/*']}

install_requires = \
['future']

setup(name='py_linq',
      version='1.2.1',
      description='Linq for Python ',
      author='Bruce Fenske',
      author_email='bwfenske@ualberta.ca',
      url='https://github.com/viralogic/py-enumerable',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
     )
